//Everything Thoks V3
//Made by Nightwolf

local enemies = {}
local pickups = {}
local scenery = {}
local springs = {}

//Scenery thok banned object flags (eth_unlimitedscenery ignores them)
local sceneryflagbans = MF_ENEMY|MF_BOSS|MF_MONITOR|MF_SPECIAL|MF_AMBIENT|MF_BOXICON

addHook("NetVars",function(net)
	enemies = net(enemies)
	pickups = net(pickups)
	scenery = net(scenery)
	springs = net(springs)
end)

//Remove all scenery restrictions except for minecarts
local et_unlimitedscenery = CV_RegisterVar({
	name = "eth_unlimitedscenery",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

local et_enemythok = CV_RegisterVar({
	name = "eth_enemythok",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

local et_bossthok = CV_RegisterVar({
	name = "eth_bossthok",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Pickup: rings and monitors
local et_pickupthok = CV_RegisterVar({
	name = "eth_pickupthok",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Player Thok - every character has thok as the main character ability
local et_playerthok = CV_RegisterVar({
	name = "eth_playerthok",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Mute sounds made by this mod
local et_mutesounds = CV_RegisterVar({
	name = "eth_mutesounds",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Springs: everything that has MF_SPRING flag or steam and fan
local et_springthok = CV_RegisterVar({
	name = "eth_springthok",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Scenery: everything else
local et_scenerythok = CV_RegisterVar({
	name = "eth_scenerythok",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//Are special stages enabled or disabled (doesn't work for NiGHTS stages)
local et_specialstage = CV_RegisterVar({
	name = "eth_specialstage",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

local function isMoPickup(mo)
	if G_IsSpecialStage(gamemap) then
		local nightsstuff = not ((mo.type >= MT_NIGHTSDRONE and mo.type <= MT_NIGHTSSTAR) or mo.type == MT_EGGCAPSULE)
		return (mo.flags & (MF_SPECIAL|MF_MONITOR) and (not (mo.flags & (MF_ENEMY|MF_BOSS|MF_FIRE))) and mo.type != MT_EGGSHIELD and mo.type != MT_STARPOST and mo.type != MT_MINECARTSPAWNER and nightsstuff)
	else
		return (mo.flags & (MF_SPECIAL|MF_MONITOR) and (not (mo.flags & (MF_ENEMY|MF_BOSS|MF_FIRE))) and mo.type != MT_EGGSHIELD and mo.type != MT_STARPOST and mo.type != MT_MINECARTSPAWNER)
	end
end

local function isMoScenery(mo)
	if et_unlimitedscenery.value == 1 then
		return true
	else
		local isbanned = (mo.flags & sceneryflagbans) or (mo.type >= MT_MACEPOINT and mo.type <= MT_BIGGRABCHAIN) or mo.type == MT_THOK or mo.type == MT_DUSTDEVIL or mo.type == MT_IVSP or mo.type == MT_SUPERSPARK or (mo.type >= MT_RAIN and mo.type <= MT_MACHINEAMBIENCE)
		return (not isbanned and mo.flags & MF_SCENERY)
	end
end

local function isMoSpring(mo)
	return ((mo.flags & MF_SPRING) or mo.type == MT_STEAM or mo.type == MT_FAN)
end

local function etplaysound(mo,sound)
	if et_mutesounds.value == 1 then return end
	S_StartSound(mo,sound)
end

//Try to add an object to one of the tables
//This function originates from an unreleased "Everything Thoks in Doom" mod (EvThoksDoom)
local function ETAddtoarray(mo)
	if mo.type == MT_PLAYER or mo.ethadded then return end
	if (mo.type >= MT_MINECART and mo.type <= MT_MINECARTENDSOLID) or mo.type == MT_MINECARTSWITCHPOINT then return end
	
	if mo.flags & MF_ENEMY or mo.flags & MF_BOSS then 
		table.insert(enemies,mo)
	elseif isMoPickup(mo) then
		table.insert(pickups,mo)
	elseif isMoSpring(mo) then
		table.insert(springs,mo)
	elseif isMoScenery(mo) then
		if not multiplayer and mapheaderinfo[gamemap].levelflags & LF_NORELOAD then //NoReload maps have no scenerythok due to unfixable bugs
			return
		else
			table.insert(scenery,mo)
		end
	else
		return
	end
	
	mo.ethadded = true
end

//Remove all dead and invalid objects in a fast way
local function ETFastMassRemove()
	local iseth = (et_enemythok.value or et_pickupthok.value or et_springthok.value or et_scenerythok.value)
	if not iseth then return end
	
	local newtable = {}
	for k,v in ipairs(enemies) do
		if v.valid and v.health > 0 then 
			table.insert(newtable,v)
		end
	end
	enemies = newtable
	
	newtable = {}
	for k,v in ipairs(pickups) do
		if v.valid and v.health > 0 then 
			table.insert(newtable,v)
		end
	end
	pickups = newtable
	
	newtable = {}
	for k,v in ipairs(springs) do
		if v.valid and v.health > 0 then 
			table.insert(newtable,v)
		end
	end
	springs = newtable
	
	newtable = {}
	for k,v in ipairs(scenery) do
		if v.valid and v.health > 0 then 
			table.insert(newtable,v)
		end
	end
	scenery = newtable
end

//Thok at the player
local function enemythok(mo)
	if not mo.valid then return end
	if mo.ethcounter == nil then mo.ethcounter = 0 end
	if not mo.target then return end
	
	if not (mo.flags & MF_NOGRAVITY) then
		if P_IsObjectOnGround(mo) and P_RandomChance(FRACUNIT/32) then
			mo.ethcounter = 1
			P_SetObjectMomZ(mo,10*FRACUNIT)
			etplaysound(mo,sfx_jump)
		elseif mo.ethcounter and mo.momz <= 0 then
			mo.ethcounter = 0
			mo.momz = 0
			P_InstaThrust(mo,R_PointToAngle2(mo.x,mo.y,mo.target.x,mo.target.y),60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		end
	else
		if not mo.ethcounter and P_RandomChance(FRACUNIT/32) then
			mo.ethcounter = 35
			P_InstaThrust(mo,R_PointToAngle2(mo.x,mo.y,mo.target.x,mo.target.y),60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		elseif mo.ethcounter then
			mo.ethcounter = $1 - 1
		end
	end
end

//Thok in a direction opposite to a player
local function pickupthok(mo)
	if not mo.valid then return end
	if mo.ethcounter == nil then mo.ethcounter = 0 end
	if not P_LookForPlayers(mo,0,true) then return end
	
	if not (mo.flags & MF_NOGRAVITY) then
		if P_IsObjectOnGround(mo) and P_RandomChance(FRACUNIT/32) then
			mo.ethcounter = 1
			P_SetObjectMomZ(mo,10*FRACUNIT)
			etplaysound(mo,sfx_jump)
		elseif mo.ethcounter and mo.momz <= 0 then
			local angle = R_PointToAngle2(mo.x,mo.y,mo.target.x,mo.target.y)+ANGLE_180
			mo.ethcounter = 0
			mo.momz = 0
			P_InstaThrust(mo,angle,60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		end
	else
		if not mo.ethcounter and P_RandomChance(FRACUNIT/32) then
			local angle = R_PointToAngle2(mo.x,mo.y,mo.target.x,mo.target.y)+ANGLE_180
			mo.ethcounter = 35
			P_InstaThrust(mo,angle,60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		elseif mo.ethcounter then
			mo.ethcounter = $1 - 1
		end
	end
end

//Thok in a random direction
local function scenerythok(mo)
	if not mo.valid then return end
	if mo.ethcounter == nil then mo.ethcounter = 0 end
	
	if not (mo.flags & MF_NOGRAVITY) then
		if P_IsObjectOnGround(mo) and P_RandomChance(FRACUNIT/32) then
			mo.ethcounter = 1
			P_SetObjectMomZ(mo,10*FRACUNIT)
			etplaysound(mo,sfx_jump)
		elseif mo.ethcounter and mo.momz <= 0 then
			local angle = FixedAngle(P_RandomRange(0,359)*FRACUNIT)
			mo.ethcounter = 0
			mo.momz = 0
			P_InstaThrust(mo,angle,60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		end
	else
		if not mo.ethcounter and P_RandomChance(FRACUNIT/32) then
			local angle = FixedAngle(P_RandomRange(0,359)*FRACUNIT)
			mo.ethcounter = 35
			P_InstaThrust(mo,angle,60*mo.scale)
			etplaysound(mo,sfx_thok)
			local thokmobj = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
			if thokmobj then thokmobj.color = SKINCOLOR_BLUE end
		elseif mo.ethcounter then
			mo.ethcounter = $1 - 1
		end
	end
end

//Try to add every spawned object to one of the tables
addHook("MobjSpawn",ETAddtoarray)

//Reset tables after a map change (NoReload maps probably use ETFastMassRemove() for that)
addHook("MapChange",function(mapnum)
	enemies = {}
	pickups = {}
	springs = {}
	scenery = {}
end)

//Main Everything Thoks frame
addHook("ThinkFrame",function()
	ETFastMassRemove() //Remove all invalid objects before doing other things
	if G_IsSpecialStage(gamemap) and et_specialstage.value == 0 then return end
	if et_enemythok.value == 1 or et_bossthok.value == 1 then
		for i = 1,#enemies do
			if (enemies[i].flags & MF_ENEMY and et_enemythok.value == 1) or (enemies[i].flags & MF_BOSS and et_bossthok.value == 1) then
				enemythok(enemies[i])
			end
		end
	end
	if et_pickupthok.value == 1 then
		for i = 1,#pickups do
			pickupthok(pickups[i])
		end
	end
	if et_springthok.value == 1 then
		for i = 1,#springs do
			scenerythok(springs[i])
		end
	end
	if et_scenerythok.value == 1 then
		for i = 1,#scenery do
			scenerythok(scenery[i])
		end
	end
end)

//eth_playerthok stuff
addHook("AbilitySpecial",function(p)
	if not p.mo then return nil end
	if p.charability != CA_THOK and et_playerthok.value == 1 then
		if not (p.pflags & PF_THOKKED) then
			P_InstaThrust(p.mo,p.mo.angle,60*p.mo.scale)
			etplaysound(p.mo,sfx_thok)
			local thokmobj = P_SpawnMobj(p.mo.x,p.mo.y,p.mo.z,MT_THOK)
			if thokmobj then thokmobj.color = p.mo.color end
			p.pflags = $1 | PF_THOKKED
		end
		return true
	end
end)